package jawful;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Vector;

/**
 * NB: only the -classpath flag flag is supported.
 */
public class CommandConverter {

	private static final String JAWFUL_HOME_SYSPROP_NAME = "jawful.jar.path";
	private static final String CLASSPATH_FILENAME_PREFIX = "jawful.classpath.";
	private static final String TIME_STAMP_FORMAT = "yyyyMMdd-HHmmss";
	private static final int CLASSPATH_LENGTH_LIMIT = 2000;
	private static final String JARFUL_JAR_NAME = "jawful.jar";
	
	/**
	 * Checks if is conversion required for the supplied command.
	 * NB: only the -classpath flag flag is supported
	 * 
	 * @param cmd the cmd
	 * 
	 * @return true if the -classpath entry is > 2000 chars.
	 */
	public boolean isConversionRequired(String[] cmd) {
		
		Vector<String> command = new Vector<String>(Arrays.asList(cmd));
		int classPathIndex = getClasspathIndex(command);
		String classpath = command.get(classPathIndex);
		
		return (classpath.length() > CLASSPATH_LENGTH_LIMIT);
	}
	
	/**
	 * Convert.
	 * NB: only the -classpath flag flag is supported
	 * 
	 * @param cmd the cmd
	 * 
	 * @return the string[]
	 */
	public String[] convert(String[] cmd) {
		Vector<String> command = new Vector<String>(Arrays.asList(cmd));

		// Swap classpath
		File jawfulJarPath = determineJawfulPath();
		
		int classPathIndex = getClasspathIndex(command);
		
		String originalClasspath = command.set(classPathIndex, jawfulJarPath.getAbsolutePath());
		
		// Write old classpath to file
		String timestamp = new SimpleDateFormat(TIME_STAMP_FORMAT).format(new Date());		
		File classPathFile = new File(CLASSPATH_FILENAME_PREFIX + timestamp);
		writeFile(classPathFile, originalClasspath);
		
		// Insert jarWrapper and classpath file
		command.insertElementAt(classPathFile.getAbsolutePath(), classPathIndex+1);
		command.insertElementAt(Wrapper.class.getCanonicalName(), classPathIndex+1);
		
		return command.toArray(new String[]{});
	}
	
	
	
	/**
	 * Determine jawful.jar path.
	 * 
	 * @return the file
	 */
	public File determineJawfulPath() {
		// First try system property
		String sysprop = System.getProperty(JAWFUL_HOME_SYSPROP_NAME);
		if (sysprop != null) {
			File jawfulJarPath = new File(sysprop);
			if (jawfulJarPath.exists()) {
				return jawfulJarPath;
			}
		} 

		// If not found, try scanning current class path
		URL[] sysUrls = ((URLClassLoader)ClassLoader.getSystemClassLoader()).getURLs();
		URL[] appUrls = ((URLClassLoader)this.getClass().getClassLoader()).getURLs();
		List<URL> urls = new ArrayList<URL>(Arrays.asList(sysUrls));
		urls.addAll(Arrays.asList(appUrls));
		
		for (URL url : urls) {
			File cpelement;
			try {
				cpelement = new File(url.toURI());
			} catch (URISyntaxException e) {
				throw new RuntimeException(e);
			}
			if (cpelement.getAbsolutePath().contains(JARFUL_JAR_NAME)) {
				return cpelement;
			}
		}

		
		// No luck - fail.
		throw new RuntimeException("Failed to find " + JARFUL_JAR_NAME + " in current classpath, and system prop "
				+ JAWFUL_HOME_SYSPROP_NAME + " wasn't set. Could not determine classpath to use when launching wrapper process.");
	}

	/**
	 * Return the index of the classpath field in the command.
	 * NB: only the -classpath flag flag is supported
	 * 
	 * @param command the command
	 * 
	 * @return the classpath index
	 */
	public int getClasspathIndex(Vector<String> command) {
		for (int i = 0; i < command.size(); i++) {
			String element = command.get(i);
			if ("-classpath".equals(element)) {
				if (command.size() < i+1) { 
					throw new RuntimeException("-classpath found but there was nothing after it");
				}
				return i + 1;
			}
		}
		throw new RuntimeException("-classpath option not found.");
	}
	
	/**
	 * Write a string to file as UTF-8. 
	 * 
	 * @param File the file to write to
	 * @param content the content to write
	 */
	public void writeFile(File file, String content) {
		
		Writer writer = null;
		try {
			writer = new OutputStreamWriter(new FileOutputStream(file), Charset.forName("UTF-8"));
			writer.write(content);
		} catch (IOException e) {
			throw new RuntimeException("Failed to write classpath [" + content + "] to file [" + file.getAbsolutePath() + "]", e);
		} finally {
			if (writer != null) {
				try { writer.close(); } catch (IOException e) {}
			}
		}
		
	}

	/**
	 * Describe command.
	 * 
	 * @param cmd
	 *            the cmd
	 * 
	 * @return the command as a string
	 */
	public String describeCommand(String[] cmd) {
		String command = "";
		for (String element : cmd) {
			command += element + " ";
		}
		return command;
	}
}
