package jawful;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

public class Wrapper {

	/**
	 * @param args class_path_file app_class [app_args]
	 * @throws Throwable the exception thrown by the wrapped class
	 */
	public static void main(String[] args) throws Throwable {
		log("In jawful.Wrapper.");
		
		Wrapper wrapper = new Wrapper();
		if (args.length < 2) {
			wrapper.usage(); // Must have at least 2 args
			return;
		}
		
		// Process arguments
		String classPath = args[0];
		String classToLaunch = args[1];
		String[] appArgs;
		if (args.length >= 3) {
			appArgs = new String[args.length-2];
			System.arraycopy(args, 2, appArgs, 0, args.length-2);
		} else { 
			appArgs = new String[] {};
		}

		// Wrap!
		try {
			wrapper.wrap(classToLaunch, classPath, appArgs);
		} catch (InvocationTargetException e) {
			log("Invocation target threw an exception. Unwrapping and re-throwing." + e);
			throw e.getCause();
		}
	}

	public void usage() {
		System.out.println("Usage:");
		System.out.println("java [jvm_args] jawful.Wrapper <class_path_file> <app_class> [app_args]");
	}

	/**
	 * Programatically load the classpath specified within file classPathFile,
	 * then launch main method of classToLaunch with the supplied args
	 * 
	 * @param classToLaunch
	 *            the class to launch
	 * @param classPathFilename
	 *            Path to file containing the classpath to load. Encoding must be UTF-8. 
	 * @param appArgs
	 *            the app args
	 * @throws Throwable 
	 */
	public void wrap(String classToLaunch, String classPathFilename, String[] appArgs) throws InvocationTargetException {
		log("Loading from: " + classPathFilename);
		
		String classpath = readClasspathFile(classPathFilename);
		
		this.addUrlsToSystemClassLoader(classpath);
		
		try {
			launchClass(classToLaunch, appArgs);			
		} catch (ClassNotFoundException e) {
			throw new RuntimeException("Jawful: failed to launch ["
					+ classToLaunch + "] with classpath [" + classPathFilename + "]:" + classpath, e);
		} catch (NoSuchMethodException e) {
			throw new RuntimeException("Jawful: failed to launch ["
					+ classToLaunch + "] with classpath [" + classPathFilename + "]:" + classpath, e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException("Jawful: failed to launch ["
					+ classToLaunch + "] with classpath [" + classPathFilename + "]:" + classpath, e);
		}
	}

	/**
	 * Read content of file and return it. If the file contains multiple lines,
	 * they will be separated by the platform path separator.
	 * 
	 * @param classPathFilename
	 *            the class path filename
	 * 
	 * @return the content of the file, with each line separated by the platform
	 *         path separator.
	 */
	public String readClasspathFile(String classPathFilename)  {
		
		File classPathFile = new File(classPathFilename);
		if (!classPathFile.exists()) {
			throw new RuntimeException("Jawful: could not find classpath file: " + classPathFilename);
		}
		
		String content;
		InputStream stream = null;
		BufferedReader bufferedReader = null;
		try {
			stream = new FileInputStream(classPathFile);
			bufferedReader = new BufferedReader(new InputStreamReader(stream , Charset.forName("UTF-8")));
			StringBuilder builder = new StringBuilder();
			String line; 
			while ((line = bufferedReader.readLine()) != null) {
				builder.append(builder.length() !=0 ? File.pathSeparator : "");				
				builder.append(line);
			}
			content = builder.toString();
			if (content == null) {
				content = "";
			}
		}  catch (IOException e) {
			throw new RuntimeException("Jawful: failed to read [" + classPathFile + "] :", e);
		} finally {
			if (stream != null) {
				try { stream.close(); } catch (IOException e) { }
			}
			if (bufferedReader != null) {
				try { bufferedReader.close(); } catch (IOException e) { }
			}
		}
		
		return content;
	}

	
	/**
	 * Launch class.
	 * 
	 * @param classToLaunch the class to launch
	 * @param appArgs the app args
	 * 
	 * @throws ClassNotFoundException the class not found exception
	 * @throws NoSuchMethodException the no such method exception
	 * @throws IllegalAccessException the illegal access exception
	 * @throws InvocationTargetException the invocation target exception
	 */
	public void launchClass(String classToLaunch, String[] appArgs) throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InvocationTargetException {
		// Look up class
		Class<?> target = Class.forName(classToLaunch);
		Class<String[]> stringArrayClass = String[].class;
		Method main = target.getMethod("main", stringArrayClass);
		
		// Now invoke the main method on that class
		main.invoke(null, (Object)appArgs);
	}

	/**
	 * Use reflection to add the classpath to the system classpath loader.
	 */
	public void addUrlsToSystemClassLoader(String classPath) {
		ClassLoader cl = ClassLoader.getSystemClassLoader();
		String addUrl = "addURL";
		if (!(cl instanceof URLClassLoader)) {
			throw new RuntimeException("");
		}
		try {
			Class<?> clazz = URLClassLoader.class;
			Class<?>[] parameterTypes = new Class[] { URL.class };
			Method addUrlMethod = clazz.getDeclaredMethod(addUrl, parameterTypes);
			addUrlMethod.setAccessible(true);

			List<String> parts = split(classPath, File.pathSeparator);
			for (String part : parts) {
				try {
					Object[] arguments = new Object[] { new File(part).toURI().toURL() };
					addUrlMethod.invoke(cl, arguments);
				} catch (Exception e) {
					log("Jawful wrapper failed to add  [" + part + "] from [" + classPath + "] to classloader: " + e);
					e.printStackTrace();
				}
			}
		} catch (NoSuchMethodException e) {
			System.err.println("Jawful failed to invoke classloader addurl method [" + cl + "]:");
			e.printStackTrace();
		}
	
	}
	
	/**
	 * Faster than using .split, according to Zero code.
	 */
	public static List<String> split(String str, String split) {
		List<String> result = new ArrayList<String>();
		String rest = str;
		while (rest != null) {
			int index = rest.indexOf(split);
			if (index == -1) {
				result.add(rest);
				rest = null;
			} else {
				result.add(rest.substring(0, index));
				if (rest.length() > index + split.length()) {
					rest = rest.substring(index + split.length());
				} else {
					rest = null;
				}
			}
		}
		return result;
	}
	
	/**
	 * Log.
	 * 
	 * @param message the message
	 */
	public static void log(String message) {
		System.out.println(message);
	}
}